import pytest

from wam2layers.config import Config

EXAMPLE_CONFIG = "example_config.yaml"


def test_load_example_config():
    Config.from_yaml(EXAMPLE_CONFIG)


def test_config_fails_for_wrong_time():
    config = Config.from_yaml(EXAMPLE_CONFIG)

    with pytest.raises(ValueError) as excinfo:
        config.preprocess_end_date = config.preprocess_start_date

    assert "End date should be later than start date" in str(excinfo.value)



import time

import psutil


class Profiler:
    def __init__(self):
        self.t0 = time.time()
        self.mem0 = self._current_memory()

    def __call__(self):
        """Report memory and time increments."""
        dt = round((time.time() - self.t0), 2)
        dmem = self._current_memory() - self.mem0
        # self.t0 = time.time()
        # self.mem0 = self._current_memory()
        return dt, dmem

    def _current_memory(self):
        return psutil.Process().memory_info().rss / (1024 * 1024)


class ProgressTracker:
    def __init__(self, output):
        """Keep track of tagged and tracked moisture."""
        self.total_tagged_moisture = 0
        self.tracked = 0
        self.boundary = 0
        self.store_intermediate_states(output)
        self.profile = Profiler()

    def store_intermediate_states(self, output):
        """Update moisture totals based on output since previous time step.

        Intended for when the output fields are reset at write time, so we
        don't lose accumulations from previous outputs.
        """
        totals = output.sum()
        self.tracked += totals["e_track"]
        self.total_tagged_moisture += totals["tagged_precip"]
        self.boundary += (
            totals["north_loss"]
            + totals["south_loss"]
            + totals["east_loss"]
            + totals["west_loss"]
        )

    def print_progress(self, t, output):
        """Print some useful runtime diagnostics."""
        totals = output.sum()
        tracked = self.tracked + totals["e_track"]
        total_tagged_moisture = self.total_tagged_moisture + totals["tagged_precip"]
        boundary = self.boundary + (
            totals["north_loss"]
            + totals["south_loss"]
            + totals["east_loss"]
            + totals["west_loss"]
        )
        still_in_atmosphere = (
            totals["s_track_upper_restart"] + totals["s_track_lower_restart"]
        )

        total_tracked_moisture = tracked + still_in_atmosphere + boundary
        tracked_percentage = (tracked + boundary) / total_tagged_moisture * 100
        lost_percentage = (1 - total_tracked_moisture / total_tagged_moisture) * 100

        time, memory = self.profile()

        print(
            f"{t} - "
            f"Tracked moisture: {tracked_percentage.item():.2f}%. "
            f"Lost moisture: {lost_percentage.item():.2f}%. "
            f"Time since start: {time}s, RAM: {memory:.2f} MB"
        )



"""Command-line interface for the WAM2layers model."""
import click

from wam2layers.analysis.visualization import cli as visualize_cli
from wam2layers.preprocessing.cli import cli as preproc_cli
from wam2layers.tracking.backtrack import cli as backtrack_cli


@click.group()
def cli():
    """Command line interface to WAM2layers."""
    pass


cli.add_command(preproc_cli, name="preprocess")
cli.add_command(backtrack_cli, name="backtrack")
cli.add_command(visualize_cli, name="visualize")



from functools import lru_cache
from pathlib import Path

import click
import numpy as np
import pandas as pd
import xarray as xr
import yaml

from wam2layers.config import Config
from wam2layers.preprocessing.shared import get_grid_info
from wam2layers.utils.profiling import Profiler, ProgressTracker


def get_tracking_dates(config):
    """Dates for tracking."""
    # E.g. if data from 00:00 to 23:00
    # Using target freq directly would end at 23:45 or so
    input_dates = pd.date_range(
        start=config.track_start_date,
        end=config.track_end_date,
        freq=config.input_frequency,
    )

    return pd.date_range(
        start=input_dates[0],
        end=input_dates[-1],
        freq=config.target_frequency,
        inclusive="right",
    )


def input_path(date, config):
    input_dir = config.preprocessed_data_folder
    return f"{input_dir}/{date.strftime('%Y-%m-%d')}_fluxes_storages.nc"


def output_path(date, config):
    output_dir = config.output_folder
    return f"{output_dir}/{date.strftime('%Y-%m-%d')}_s_track.nc"


# LRU Cache keeps the file open so we save a bit on I/O
@lru_cache(maxsize=2)
def read_data_at_date(d):
    """Load input data for given date."""
    file = input_path(d, config)
    return xr.open_dataset(file, cache=False)


# This one can't be cached as we'll be overwriting the content.
def read_data_at_time(t):
    """Get a single time slice from input data at time t."""
    ds = read_data_at_date(t)
    return ds.sel(time=t, drop=True)


def load_data(t, subset="fluxes"):
    """Load variable at t, interpolate if needed."""
    variables = {
        "fluxes": ["fx_upper", "fx_lower", "fy_upper", "fy_lower", "evap", "precip"],
        "states": ["s_upper", "s_lower"],
    }

    t1 = t.ceil(config.input_frequency)
    da1 = read_data_at_time(t1)[variables[subset]]
    if t == t1:
        # this saves a lot of work if times are aligned with input
        return da1

    t0 = t.floor(config.input_frequency)
    da0 = read_data_at_time(t0)[variables[subset]]
    if t == t0:
        return da0

    return da0 + (t - t0) / (t1 - t0) * (da1 - da0)


def load_fluxes(t):
    t_current = t - pd.Timedelta(config.target_frequency) / 2
    return load_data(t_current, "fluxes")


def load_states(t):
    t_prev = t
    t_next = t - pd.Timedelta(config.target_frequency)
    states_prev = load_data(t_prev, "states")
    states_next = load_data(t_next, "states")
    return states_prev, states_next


def time_in_range(start, end, current):
    """Returns whether current is in the range [start, end]"""
    return start <= current <= end


def load_region(config):
    # TODO: make variable name more generic
    return xr.open_dataset(config.region).source_region


def to_edges_zonal(fx, periodic_boundary=False):
    """Define the horizontal fluxes over the east/west boundaries."""
    fe = np.zeros_like(fx)
    fe[:, :-1] = 0.5 * (fx[:, :-1] + fx[:, 1:])
    if periodic_boundary:
        fe[:, -1] = 0.5 * (fx[:, -1] + fx[:, 0])

    # find out where the positive and negative fluxes are
    f_pos = np.ones_like(fx)
    f_pos[fe < 0] = 0
    f_neg = f_pos - 1

    # separate directions west-east (all positive numbers)
    fe_we = fe * f_pos
    fe_ew = fe * f_neg

    # fluxes over the western boundary
    fw_we = look_west(fe_we)
    fw_ew = look_west(fe_ew)

    return fe_we, fe_ew, fw_we, fw_ew


def to_edges_meridional(fy):
    """Define the horizontal fluxes over the north/south boundaries."""
    fn = np.zeros_like(fy)
    fn[1:, :] = 0.5 * (fy[:-1, :] + fy[1:, :])

    # find out where the positive and negative fluxes are
    fn_pos = np.ones_like(fn)
    fn_pos[fn < 0] = 0  # invalid value encountered in less
    fn_neg = fn_pos - 1

    # separate directions south-north (all positive numbers)
    fn_sn = fn * fn_pos
    fn_ns = fn * fn_neg

    # fluxes over the southern boundary
    fs_sn = look_south(fn_sn)
    fs_ns = look_south(fn_ns)

    return fn_sn, fn_ns, fs_sn, fs_ns


def look_north(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, 1, axis=-2)


def look_south(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, -1, axis=-2)


def look_east(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, -1, axis=-1)


def look_west(array):
    # Note: edges are reinserted at other end; but they're not used anyway
    return np.roll(array, 1, axis=-1)


def split_vertical_flux(Kvf, fv):
    f_downward = np.zeros_like(fv)
    f_upward = np.zeros_like(fv)
    f_downward[fv >= 0] = fv[fv >= 0]
    f_upward[fv <= 0] = fv[fv <= 0]
    f_upward = np.abs(f_upward)

    # include the vertical dispersion
    if Kvf != 0:
        f_upward = (1.0 + Kvf) * f_upward
        f_upward[fv >= 0] = fv[fv >= 0] * Kvf
        f_downward = (1.0 + Kvf) * f_downward
        f_downward[fv <= 0] = np.abs(fv[fv <= 0]) * Kvf

    return f_downward, f_upward


def change_units(data, target_freq):
    """Change units to m3.
    Multiply by edge length or area to get flux in m3
    Multiply by time to get accumulation instead of flux
    Divide by density of water to go from kg to m3
    """
    density = 1000  # [kg/m3]
    a, ly, lx = get_grid_info(data)

    total_seconds = pd.Timedelta(target_freq).total_seconds()

    for variable in data.data_vars:
        if variable in ["fx_upper", "fx_lower"]:
            data[variable] *= total_seconds / density * ly
        elif variable in ["fy_upper", "fy_lower"]:
            data[variable] *= total_seconds / density * lx[:, None]
        elif variable in ["evap", "precip"]:
            data[variable] *= total_seconds / density * a[:, None]
        elif variable in ["s_upper", "s_lower"]:
            data[variable] *= a[:, None] / density
        else:
            raise ValueError(f"Unrecognized variable {variable}")
        data[variable] = data[variable].assign_attrs(units="m**3")


def stabilize_fluxes(current, previous):
    """Stabilize the outfluxes / influxes.

    CFL: Water cannot move further than one grid cell per timestep.
    """
    for level in ["upper", "lower"]:
        fx = current["fx_" + level]
        fy = current["fy_" + level]
        s = previous["s_" + level]

        fx_abs = np.abs(fx)
        fy_abs = np.abs(fy)
        ft_abs = fx_abs + fy_abs

        fx_corrected = 1 / 2 * fx_abs / ft_abs * s.values
        fx_stable = np.minimum(fx_abs, fx_corrected)

        fy_corrected = 1 / 2 * fy_abs / ft_abs * s.values
        fy_stable = np.minimum(fy_abs, fy_corrected)

        # Get rid of any nan values
        fx_stable.fillna(0)
        fy_stable.fillna(0)

        # Re-instate the sign
        current["fx_" + level] = np.sign(fx) * fx_stable
        current["fy_" + level] = np.sign(fy) * fy_stable


def convergence(fx, fy):
    # Note: latitude decreasing, hence positive fy gradient is convergence
    return np.gradient(fy, axis=-2) - np.gradient(fx, axis=-1)


def calculate_fv(fluxes, states_prev, states_next):
    """Calculate the vertical fluxes.

    Note: fluxes are given at temporal midpoints between states.
    """
    s_diff = states_prev - states_next
    s_mean = (states_prev + states_next) / 2
    s_total = s_mean.s_upper + s_mean.s_lower
    s_rel = s_mean / s_total

    tendency_upper = (
        convergence(fluxes.fx_upper, fluxes.fy_upper)
        - fluxes.precip.values * s_rel.s_upper
    )
    tendency_lower = (
        convergence(fluxes.fx_lower, fluxes.fy_lower)
        - fluxes.precip.values * s_rel.s_lower
        + fluxes.evap
    )

    residual_upper = s_diff.s_upper - tendency_upper
    residual_lower = s_diff.s_lower - tendency_lower

    # compute the resulting vertical moisture flux; the vertical velocity so
    # that the new residual_lower/s_lower = residual_upper/s_upper (positive downward)
    fv = s_rel.s_lower * (residual_upper + residual_lower) - residual_lower

    # stabilize the outfluxes / influxes; during the reduced timestep the
    # vertical flux can maximally empty/fill 1/x of the top or down storage
    stab = 1.0 / (config.kvf + 1.0)
    flux_limit = np.minimum(s_mean.s_upper, s_mean.s_lower)
    fv_stable = np.minimum(np.abs(fv), stab * flux_limit)

    # Reinstate the sign
    return np.sign(fv) * fv_stable


def backtrack(
    fluxes,
    states_prev,
    states_next,
    region,
    output,
):

    # Unpack input data
    fx_upper = fluxes["fx_upper"].values
    fy_upper = fluxes["fy_upper"].values
    fx_lower = fluxes["fx_lower"].values
    fy_lower = fluxes["fy_lower"].values
    evap = fluxes["evap"].values
    precip = fluxes["precip"].values
    f_vert = fluxes["f_vert"].values
    s_upper = states_prev["s_upper"].values
    s_lower = states_prev["s_lower"].values

    s_track_upper = output["s_track_upper_restart"].values
    s_track_lower = output["s_track_lower_restart"].values

    tagged_precip = region * precip
    s_total = s_upper + s_lower

    # separate the direction of the vertical flux and make it absolute
    f_downward, f_upward = split_vertical_flux(config.kvf, f_vert)

    # Determine horizontal fluxes over the grid-cell boundaries
    f_e_lower_we, f_e_lower_ew, f_w_lower_we, f_w_lower_ew = to_edges_zonal(
        fx_lower, config.periodic_boundary
    )
    f_e_upper_we, f_e_upper_ew, f_w_upper_we, f_w_upper_ew = to_edges_zonal(
        fx_upper, config.periodic_boundary
    )

    (
        fy_n_lower_sn,
        fy_n_lower_ns,
        fy_s_lower_sn,
        fy_s_lower_ns,
    ) = to_edges_meridional(fy_lower)
    (
        fy_n_upper_sn,
        fy_n_upper_ns,
        fy_s_upper_sn,
        fy_s_upper_ns,
    ) = to_edges_meridional(fy_upper)

    # Short name for often used expressions
    s_track_relative_lower = s_track_lower / s_lower
    s_track_relative_upper = s_track_upper / s_upper
    if config.periodic_boundary:
        inner = np.s_[1:-1, :]
    else:
        inner = np.s_[1:-1, 1:-1]

    # Actual tracking (note: backtracking, all terms have been negated)
    s_track_lower[inner] += (
        +f_e_lower_we * look_east(s_track_relative_lower)
        + f_w_lower_ew * look_west(s_track_relative_lower)
        + fy_n_lower_sn * look_north(s_track_relative_lower)
        + fy_s_lower_ns * look_south(s_track_relative_lower)
        + f_upward * s_track_relative_upper
        - f_downward * s_track_relative_lower
        - fy_s_lower_sn * s_track_relative_lower
        - fy_n_lower_ns * s_track_relative_lower
        - f_e_lower_ew * s_track_relative_lower
        - f_w_lower_we * s_track_relative_lower
        + tagged_precip * (s_lower / s_total)
        - evap * s_track_relative_lower
    )[inner]

    s_track_upper[inner] += (
        +f_e_upper_we * look_east(s_track_relative_upper)
        + f_w_upper_ew * look_west(s_track_relative_upper)
        + fy_n_upper_sn * look_north(s_track_relative_upper)
        + fy_s_upper_ns * look_south(s_track_relative_upper)
        + f_downward * s_track_relative_lower
        - f_upward * s_track_relative_upper
        - fy_s_upper_sn * s_track_relative_upper
        - fy_n_upper_ns * s_track_relative_upper
        - f_w_upper_we * s_track_relative_upper
        - f_e_upper_ew * s_track_relative_upper
        + tagged_precip * (s_upper / s_total)
    )[inner]

    # down and top: redistribute unaccounted water that is otherwise lost from the sytem
    lower_to_upper = np.maximum(0, s_track_lower - states_next["s_lower"])
    upper_to_lower = np.maximum(0, s_track_upper - states_next["s_upper"])
    s_track_lower[inner] = (s_track_lower - lower_to_upper + upper_to_lower)[inner]
    s_track_upper[inner] = (s_track_upper - upper_to_lower + lower_to_upper)[inner]

    # Update output fields
    output["e_track"] += evap * (s_track_lower / s_lower)
    output["north_loss"] += (
        fy_n_upper_ns * s_track_relative_upper + fy_n_lower_ns * s_track_relative_lower
    )[1, :]
    output["south_loss"] += (
        fy_s_upper_sn * s_track_relative_upper + fy_s_lower_sn * s_track_relative_lower
    )[-2, :]

    if config.periodic_boundary == False:
        output["east_loss"] += (
            f_e_upper_ew * s_track_relative_upper
            + f_e_lower_ew * s_track_relative_lower
        )[:, -2]
        output["west_loss"] += (
            f_w_upper_we * s_track_relative_upper
            + f_w_lower_we * s_track_relative_lower
        )[:, 1]

    output["s_track_lower_restart"].values = s_track_lower
    output["s_track_upper_restart"].values = s_track_upper
    output["tagged_precip"] += tagged_precip


def initialize(config_file):
    """Read config, region, and initial states."""
    print(f"Initializing experiment with config file {config_file}")

    config = Config.from_yaml(config_file)
    region = load_region(config)

    output = initialize_outputs(region)
    region = region.values

    if config.restart:
        # Reload last state from existing output
        tracking_dates = get_tracking_dates(config)
        date = tracking_dates[-1] + pd.Timedelta(days=1)
        ds = xr.open_dataset(output_path(date, config))
        output["s_track_upper_restart"].values = ds.s_track_upper_restart.values
        output["s_track_lower_restart"].values = ds.s_track_lower_restart.values

    print(f"Output will be written to {config.output_folder}.")
    return config, region, output


def initialize_outputs(region):
    """Allocate output arrays."""

    proto = region
    output = xr.Dataset(
        {
            # Keep last state for a restart
            "s_track_upper_restart": xr.zeros_like(proto),
            "s_track_lower_restart": xr.zeros_like(proto),
            "e_track": xr.zeros_like(proto),
            "tagged_precip": xr.zeros_like(proto),
            "north_loss": xr.zeros_like(proto.isel(latitude=0, drop=True)),
            "south_loss": xr.zeros_like(proto.isel(latitude=0, drop=True)),
            "east_loss": xr.zeros_like(proto.isel(longitude=0, drop=True)),
            "west_loss": xr.zeros_like(proto.isel(longitude=0, drop=True)),
        }
    )
    return output


def write_output(output, t):
    # TODO: add back (and cleanup) coordinates and units
    path = output_path(t, config)
    print(f"{t} - Writing output to file {path}")
    output.astype("float32").to_netcdf(path)

    # Flush previous outputs
    output[
        [
            "e_track",
            "tagged_precip",
            "north_loss",
            "south_loss",
            "east_loss",
            "west_loss",
        ]
    ] *= 0


#############################################################################
# With the correct import statements, the code in the function below could
# alternatively be be used as a script in a separate python file or notebook.
#############################################################################


def run_experiment(config_file):
    """Run a backtracking experiment from start to finish."""
    global config

    config, region, output = initialize(config_file)

    tracking_dates = get_tracking_dates(config)

    progress_tracker = ProgressTracker(output)
    for t in reversed(tracking_dates):

        fluxes = load_fluxes(t)
        states_prev, states_next = load_states(t)

        # Convert data to volumes
        change_units(states_prev, config.target_frequency)
        change_units(states_next, config.target_frequency)
        change_units(fluxes, config.target_frequency)

        # Apply a stability correction if needed
        stabilize_fluxes(fluxes, states_next)

        # Determine the vertical moisture flux
        fluxes["f_vert"] = calculate_fv(fluxes, states_prev, states_next)

        # Only track the precipitation at certain timesteps
        if (
            time_in_range(
                config.event_start_date,
                config.event_end_date,
                t,
            )
            == False
        ):
            fluxes["precip"] = 0

        backtrack(
            fluxes,
            states_prev,
            states_next,
            region,
            output,
        )

        # Daily output
        if t == t.floor(config.output_frequency):
            progress_tracker.print_progress(t, output)
            progress_tracker.store_intermediate_states(output)
            write_output(output, t)


###########################################################################
# The code below makes it possible to run wam2layers from the command line:
# >>> python backtrack.py path/to/cases/era5_2021.yaml
# or even:
# >>> wam2layers backtrack path/to/cases/era5_2021.yaml
###########################################################################


@click.command()
@click.argument("config_file", type=click.Path(exists=True))
def cli(config_file):
    """Run WAM2layers backtrack experiment from the command line.

    CONFIG_FILE: Path to WAM2layers experiment configuration file.

    Usage examples:

        \b
        - python path/to/backtrack.py path/to/cases/era5_2021.yaml
        - wam2layers backtrack path/to/cases/era5_2021.yaml
    """
    print("Welcome to WAM2layers.")
    print("Starting backtrack experiment.")
    run_experiment(config_file)


if __name__ == "__main__":
    cli()



from pathlib import Path

import click
import matplotlib.pyplot as plt
import pandas as pd
import xarray as xr
from cmocean import cm

from wam2layers.config import Config
from wam2layers.preprocessing.shared import get_grid_info
from wam2layers.tracking.backtrack import input_path, load_region, output_path


def try_import_cartopy():
    """Import cartopy if it is available; else raise."""
    from importlib import import_module

    global crs
    global cfeature

    try:
        crs = import_module("cartopy.crs")
        cfeature = import_module("cartopy.feature")
    except ImportError as exec:
        message = "This function requires cartopy. Cartopy is most easily installed with conda. Please refer to the documentation."
        raise ImportError(message) from exec


def polish(ax, region):
    ax.add_feature(cfeature.COASTLINE, linewidth=0.8)
    ax.add_feature(cfeature.BORDERS, linestyle="-", linewidth=0.2)
    ax.set_xlim(region.longitude.min(), region.longitude.max())
    ax.set_ylim(region.latitude.min(), region.latitude.max())


def _plot_precip(config, ax):
    """Return subplot with precip."""
    # Load config and some usful stuf.
    region = load_region(config)

    # Load data
    dates = pd.date_range(
        start=config.preprocess_start_date,
        end=config.preprocess_end_date,
        freq=config[
            "output_frequency"
        ],  # Should be output frequency, since this is used to save the data
        inclusive="left",
    )

    input_files = []
    for date in dates:
        input_files.append(input_path(date, config))

    ds = xr.open_mfdataset(input_files, combine="nested", concat_dim="time")
    # TODO: make region time-dependent
    start = config.event_start_date
    end = config.event_end_date
    subset = ds.precip.sel(time=slice(start, end))
    precip = (subset * region * 3600).sum("time").compute()

    # Make figure
    precip.plot(ax=ax, cmap=cm.rain, cbar_kwargs=dict(fraction=0.05, shrink=0.5))
    ax.set_title("Cumulative precipitation during event [mm]", loc="left")
    polish(ax, region.where(region > 0, drop=True))


def _plot_evap(config, ax):
    """Return subplot with tracked evaporation."""
    region = load_region(config)
    a_gridcell, lx, ly = get_grid_info(region)

    # Load data
    dates = pd.date_range(
        start=config.track_start_date,
        end=config.track_end_date,
        freq=config.output_frequency,
        inclusive="left",
    )

    output_files = []
    for date in dates:
        output_files.append(output_path(date, config))

    ds = xr.open_mfdataset(output_files, combine="nested", concat_dim="time")
    e_track = ds.e_track.sum("time").compute() * 1000 / a_gridcell[:, None]

    # Make figure
    e_track.plot(
        ax=ax, vmin=0, cmap=cm.rain, cbar_kwargs=dict(fraction=0.05, shrink=0.5)
    )
    e_track.plot.contour(ax=ax, levels=[0.1, 1], colors=["lightgrey", "grey"])
    ax.set_title("Moisture source of extreme precip [mm]", loc="left")

    # Add source region outline
    region.plot.contour(ax=ax, levels=[1], colors="k")
    polish(ax, region)


def visualize_input_data(config_file):
    """An figure showing the cumulative moisture inputs.

    TODO: make figure creation independent of case.
    """
    config = Config.from_yaml(config_file)

    # Make figure
    fig = plt.figure(figsize=(16, 10))
    ax = fig.add_subplot(111, projection=crs.PlateCarree())

    _plot_precip(config, ax)

    # Save
    out_dir = Path(config.output_folder) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)
    fig.savefig(out_dir / "input_event.png", dpi=200)


def visualize_output_data(config_file):
    """An figure showing the cumulative moisture origins.

    TODO: make figure creation independent of case.
    """
    # Load config and some usful stuf.
    config = Config.from_yaml(config_file)

    # Make figure
    fig = plt.figure(figsize=(16, 10))
    ax = fig.add_subplot(111, projection=crs.PlateCarree())

    _plot_evap(config, ax)

    # Save
    out_dir = Path(config.output_folder) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)
    fig.savefig(out_dir / "cumulative_sources.png", dpi=200)


def visualize_both(config_file):
    """Diagnostic figure with four subplots combining input and output data."""
    # Load config and some usful stuf.
    config = Config.from_yaml(config_file)

    # Make figure
    fig, [ax1, ax2] = plt.subplots(
        2, 1, figsize=(16, 10), subplot_kw=dict(projection=crs.PlateCarree())
    )

    _plot_precip(config, ax1)
    _plot_evap(config, ax2)

    # Save
    out_dir = Path(config.output_folder) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)
    fig.savefig(out_dir / "summary_subplots.png", dpi=200)


def visualize_snapshots(config_file):
    """Diagnostic figure with four subplots combining input and output data."""
    config = Config.from_yaml(config_file)
    dates = pd.date_range(
        start=config.track_start_date,
        end=config.track_end_date,
        freq=config.output_frequency,
        inclusive="left",
    )
    region = load_region(config)
    a_gridcell, lx, ly = get_grid_info(region)

    out_dir = Path(config.output_folder) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)

    for date in dates:
        print(date)
        ds_in = xr.open_dataset(input_path(date, config))
        ds_out = xr.open_dataset(output_path(date, config))

        # Make figure
        fig, [[ax1, ax2], [ax3, ax4]] = plt.subplots(
            2, 2, subplot_kw=dict(projection=crs.PlateCarree()), figsize=(14, 8)
        )

        ax1.set_title("Tracked precipitation" + date.strftime("%Y%m%d"))
        precip = ds_in.precip.sum("time") * region / a_gridcell[:, None] * 1000
        precip.plot(ax=ax1, cmap="Blues")
        polish(ax1, region)

        ax2.set_title("Moisture source")
        e_track = ds_out.e_track / a_gridcell[:, None] * 1000
        e_track.plot(ax=ax2, cmap="GnBu")
        polish(ax2, region)

        ax3.set_title("S track upper layer")
        s_track_upper = ds_out.s_track_upper_restart / a_gridcell[:, None] * 1000
        s_track_upper.plot(ax=ax3, cmap="YlOrRd")
        ds_in.mean("time").plot.streamplot(
            x="longitude",
            y="latitude",
            u="fx_upper",
            v="fy_upper",
            ax=ax3,
            color="black",
        )
        polish(ax3, region)

        ax4.set_title("S track lower layer")
        s_track_lower = ds_out.s_track_lower_restart / a_gridcell[:, None] * 1000
        s_track_lower.plot(ax=ax4, cmap="YlOrRd")
        ds_in.mean("time").plot.streamplot(
            x="longitude",
            y="latitude",
            u="fx_lower",
            v="fy_lower",
            ax=ax4,
            color="black",
        )
        polish(ax4, region)

        # Save
        output_file = out_dir / f"input_output_{date.strftime('%Y%m%d')}.png"
        fig.savefig(output_file)
        plt.close()


###########################################################################
# The code below makes it possible to run wam2layers from the command line:
# >>> python visualization.py input path/to/cases/era5_2021.yaml
# or even:
# >>> wam2layers visualize output path/to/cases/era5_2021.yaml
###########################################################################


@click.group()
def cli():
    """Visualize input or output data of a WAM2layers experiment"""
    pass


@cli.command()
@click.argument("config_file", type=click.Path(exists=True))
def input(config_file):
    """Visualize input data for experiment."""
    try_import_cartopy()
    visualize_input_data(config_file)


@cli.command()
@click.argument("config_file", type=click.Path(exists=True))
def output(config_file):
    """Visualize output data for experiment."""
    try_import_cartopy()
    visualize_output_data(config_file)


@cli.command()
@click.argument("config_file", type=click.Path(exists=True))
def both(config_file):
    """Visualize both input and output data for experiment."""
    try_import_cartopy()
    visualize_both(config_file)


@cli.command()
@click.argument("config_file", type=click.Path(exists=True))
def snapshots(config_file):
    """Visualize input and output snapshots for experiment."""
    try_import_cartopy()
    visualize_snapshots(config_file)


if __name__ == "__main__":
    cli()



from pydantic import BaseModel, FilePath, validator, root_validator
import yaml
from datetime import datetime
from pathlib import Path
from typing import Literal, Union

import yaml
from pydantic import BaseModel, FilePath, validator


class Config(BaseModel):
    filename_template: str
    """The filename pattern of the raw input data.

    Used to find the input files during preprocessing. The pattern will be
    interpreted during execution of the model to find the input data for each
    date and variable.

    For example, the following pattern:

    .. code-block:: yaml

        filename_template: /ERA5data/{year}/{month:02}/ERA5_{year}-{month:02d}-{day:02d}{levtype}_{variable}.nc

    will be converted to

    .. code-block:: python

        /ERA5data/2021/07/ERA5_2021-07-15_ml_u.nc

    for date 2022-07-15, variable u, and levtype "_ml" (note the underscore).
    """

    preprocessed_data_folder: Path
    """Location where the pre-processed data should be stored.

    If it does not exist, it will be created during pre-processing.

    For example:

    .. code-block:: yaml

        preprocessed_data_folder: ~/floodcase_202107/preprocessed_data

    """

    region: FilePath
    """Location where the region mask is stored.

    The file should exist.

    For example:

    .. code-block:: yaml

        region: /data/volume_2/era5_2021/source_region_global.nc

    """

    output_folder: Path
    """Location where output of tracking and analysis should be written.

    For example:

    .. code-block:: yaml

        output_folder: ~/floodcase_202107/output_data

    """

    preprocess_start_date: datetime
    """Start date for preprocessing.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date.
    The preprocess_start_date is included in the preprocessing.

    For example:

    .. code-block:: yaml

        preprocess_start_date: "2021-07-01T00:00"

    """

    preprocess_end_date: datetime
    """End date for preprocessing.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date.
    The preprocess_end_date is included in the preprocessing.

    For example:

    .. code-block:: yaml

        preprocess_end_date: "2021-07-15T23:00"
    """

    track_start_date: datetime
    """Start date for tracking.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.
    When backward tracking the track_start_date is not given as output date.

    For example:

    .. code-block:: yaml

        track_start_date: "2021-07-01T00:00"

    """

    track_end_date: datetime
    """Start date for tracking.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.

    For example:

    .. code-block:: yaml

        track_end_date: "2021-07-15T23:00"
    """

    event_start_date: datetime
    """Start date for event.

    For tracking individual (e.g. heavy precipitation) events, you can set the
    start and end date of the event to something different than the total
    tracking start and end date, you can also indicate the hours that you want to track.
    The event_start_date is included in the event tracking.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.

    For example:

    .. code-block:: yaml

        event_start_date: "2021-07-13T00:00"

    """

    event_end_date: datetime
    """Start date for event.

    For tracking individual (e.g. heavy precipitation) events, you can set the
    start and end date of the event to something different than the total
    tracking start and end date, you can also indicate the hours that you want to track.
    The event_end_date is included in the event tracking.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.

    For example:

    .. code-block:: yaml

        event_end_date: "2021-07-14T23:00"

    """

    input_frequency: str
    """Frequency of the raw input data.

    Used to calculated water volumes.

    For example:

    .. code-block:: yaml

        input_frequency: '1h'

    """

    target_frequency: str
    """Frequence at which to perform the tracking.

    The data will be interpolated during model execution. Too low frequency will
    violate CFL criterion, too high frequency will lead to excessive numerical
    diffusion and slow progress. For best performance, the input frequency
    should be divisible by the target frequency.

    For example:

    .. code-block:: yaml

        target_frequency: '10min'

    """

    output_frequency: str
    """Frequency at which to write output to file.

    For example, for daily output files:

    .. code-block:: yaml

        output_frequency: '1d'

    """

    level_type: Literal["model_levels", "pressure_levels"]
    """Type of vertical levels in the raw input data.

    Can be either `model_levels` or `pressure_levels`.

    For example:

    .. code-block:: yaml

        level_type: model_levels

    """

    levels: "Union[list[int], Literal['All']]"
    """Which levels to use from the raw input data.

    A list of integers corresponding to the levels in the input data, or a
    subset thereof. Shorthand `"all"` will attempt to use all 137 ERA5 levels.

    For example:

    .. code-block:: yaml

        levels: [20,40,60,80,90,95,100,105,110,115,120,123,125,128,130,131,132,133,134,135,136,137]

    """

    log_level: Literal["debug", "info", "warning", "error", "critical"]
    """Verbosity of the output messages.

    For example:

    .. code-block:: yaml

        log_level: info
    """

    restart: bool
    """Whether to restart from previous run.

    If set to `true`, this will attempt to read the output from a previous model
    run and continue from there. The output from the previous timestep must be
    available for this to work.

    For example:

    .. code-block:: yaml

        restart: False

    """

    periodic_boundary: bool
    """Whether to use period boundaries in the zonal direction.

    This should be used when working with global datasets.

    For example:

    .. code-block:: yaml

        periodic_boundary: true
    """

    kvf: int
    """net to gross vertical flux multiplication parameter

    For example:

    .. code-block:: yaml

        kvf: 3

    """

    chunks: "Union[None, dict[str, int]]"
    """Whether to use dask.

    Using dask can help process large datasets without memory issues, but its
    performance is quite sensitive to the chunk configuration.

    Some examples:

    .. code-block:: yaml

        # don't use dask:
        chunks: null

        # process one time step at a time
        chunks: {time: 1}

        # set chunking for all individual dims
        chunks:
            level: -1
            time: 1  # don't set time to auto, as it will be different for surface and 3d vars
            latitude: auto
            longitude: auto

    """

    @classmethod
    def from_yaml(cls, config_file):
        """Read settings from a configuration.yaml file.

        For example:

        .. highlight:: python
        .. code-block:: python

            from wam2layers.config import Config
            config = Config.from_yaml('../../cases/floodcase_2021.yaml')

        """
        with open(config_file) as f:
            settings = yaml.safe_load(f)

        return cls(**settings)

    @validator("preprocessed_data_folder", "region", "output_folder")
    def _expanduser(cls, path):
        """Resolve ~ in input paths."""
        return path.expanduser()

    @validator("preprocessed_data_folder", "output_folder")
    def _make_dir(cls, path):
        """Create output dirs if they don't exist yet."""
        if not path.exists():
            print(f"Creating output folder {path}")
            path.mkdir(parents=True)
        return path

    @root_validator
    def check_date_order(cls, values):
        for date_field in ['track', 'event', 'preprocess']:
            start = values.get(f'{date_field}_start_date')
            end = values.get(f'{date_field}_end_date')

            if not start < end:
                raise ValueError("End date should be later than start date.")

        return values

    class Config:
        """This is the config of the **Pydantic** class, not the wam2layers config.

        Note: this will change in v2
        https://docs.pydantic.dev/blog/pydantic-v2-alpha/#changes-to-config
        """
        # Also force validation when new value is set on existing config
        validate_assignment = True

